/* @pjs preload="Modis.jpg"; 
 */

// Contrail simulator
// (c) 2011 Mick West, contrailscience.com

// distances are in meters
// times are in seconds
// speeds are in m/s  (1 mph = 0.44704 m/s)

// Deployment notes:
// Use File -> Export
// This will save to /applet
// Then upload the entire /racetrack_contrail_simulator folder to contrailscience.com/apps

String app_path = "/apps/racetrack_contrail_simulator/";

PImage backgroundImage;

float lifetime = 500;

float start_radius = 1;
float radius_grow = 0.001;

float mph_to_ms = 0.44704;
float miles_to_meters = 1609.344;

float meters_per_pixel = 250.0;

float time_between_planes = 120;
float time_until_next_plane = 0;  // 0 so we get a plane immediately.
boolean use_AB = true;              // toggle between AB and CD


//  PVector P1 = new PVector(400*miles_to_meters,100*miles_to_meters);    // first point
PVector P1 = new PVector(50*miles_to_meters,24*miles_to_meters);    // first point
PVector P2 = new PVector(73*miles_to_meters,25*miles_to_meters);     

float wind_speed_x,wind_speed_y;

float plane_speed, wind_speed, wind_angle, turn_rate, base_circumfrence,base_radius, angle;
PVector centerline,base;
float time = 0;       // current time
float time_step = 10;      // time step in seconds per update fra

// global vars that are calculated in setup
// some bug prevents us using functions like mag in this global initialization

float leg_length;
float total_length;
float leg_time;
float total_time=1;


float  old_center_x;
float  old_center_y;




boolean grid_mode = false; 


MUIElementManager MUIMan;
MPlaneManager MPlaneMan;

MUIVSlider speedSlider,windSlider,headingSlider,turnrateSlider;

MUICheckBox pauseCheckBox, dirCheckBox, fig8CheckBox, gridCheckBox;

MUIDragable P1Drag, P2Drag, CenterDrag;
MUIDragable mark_left_start, mark_left_end, mark_right_start, mark_right_end;


MUIDragable ADrag,BDrag,CDrag,DDrag;


// Helper function to to load an image, accounting for difference in path with processing.js
PImage MUILoadImage(String path)
{
  PImage i = loadImage(path);
  if (i==null || i.width <= 0)
  {
    i = loadImage(app_path+"data/"+path);
  }
  return i;
}

void setup() {
  
  backgroundImage = MUILoadImage("Modis.jpg");

  // Graphics mode setup
  size(640, 640, JAVA2D);
  smooth();
  colorMode(RGB, 255, 255, 255, 100);

  // Setup the particle system
  ps = new ParticleSystem(1000, new PVector(width/2,height/2,0));

  MPlaneMan = new MPlaneManager();

  // MUISlider Manager
  MUIMan = new MUIElementManager();

  speedSlider =      MUIMan.addVSlider(450,400,20,200,100,1000,478,1,"Plane\nSpeed\nMPH");
  windSlider =       MUIMan.addVSlider(500,400,20,200,0,100,44,1,"Wind\nSpeed\nMPH");
  headingSlider =    MUIMan.addVSlider(550,400,20,200,0,360,59,1,"Wind\nHeading\nDeg");
  turnrateSlider =   MUIMan.addVSlider(600,400,20,200,60,1200,360,10,"Turn\nrate\nsecs");
  
  pauseCheckBox = MUIMan.addCheckBox(450,620,20,20,0,"Pause");
  dirCheckBox =   MUIMan.addCheckBox(520,620,20,20,0,"Clockwise");
  fig8CheckBox = MUIMan.addCheckBox(380,620,20,20,0,"Fig8");  
  gridCheckBox = MUIMan.addCheckBox(20,620,20,20,0,"Intersection");
  
  P1Drag = MUIMan.addDragable(50*miles_to_meters/meters_per_pixel,24*miles_to_meters/meters_per_pixel,20,20,0,"P1");
  P2Drag = MUIMan.addDragable(73*miles_to_meters/meters_per_pixel,25*miles_to_meters/meters_per_pixel,20,20,0,"P2");
  
  // AB is a line from the left to the right
  ADrag = MUIMan.addDragable(0,50,20,10,0,"A");  
  BDrag = MUIMan.addDragable(620,150,20,10,0,"B");
  
  // CD is one from the top to the bottom
  CDrag = MUIMan.addDragable(500,0,10,20,0,"C");  
  DDrag = MUIMan.addDragable(150,620,10,20,0,"D");
  
  
  /*
  // some debug stuff
  mark_left_start = MUIMan.addDragable(50,50,10,10,0,"LS");
  mark_left_end = MUIMan.addDragable(0,0,10,10,0,"LE");
  mark_right_start = MUIMan.addDragable(0,0,10,10,0,"RS");
  mark_right_end = MUIMan.addDragable(0,0,10,10,0,"RE");
  */
  
  CenterDrag = MUIMan.addDragable((P1Drag.x+P2Drag.x)/2,(P1Drag.y+P2Drag.y)/2,10,10,0,"");
  
  
  gridCheckBox.value=1;
  
}

void Pline(PVector A, PVector B)
{
  line(A.x, A.y, B.x, B.y);
}

int drawTime = 0;  //for fps
int calcFPS(){
  int diff = millis() - drawTime;
  int fps = (int)(1000/ float(diff));
  drawTime = millis();
  return fps;
} 

void drawRacetrack() {

        MUIMan.display();


  /*  
   slider1.display();
   if(slider1.check(mouseX,mouseY))
   {  
   slider1.move(mouseY);
   // sine1.setAmp(slideramp1.valeur);
   }
   */


  
  if (CenterDrag.x != old_center_x || CenterDrag.y != old_center_y)
  {
    P1Drag.x += CenterDrag.x - old_center_x;
    P1Drag.y += CenterDrag.y - old_center_y;
    P2Drag.x += CenterDrag.x - old_center_x;
    P2Drag.y += CenterDrag.y - old_center_y;
  }
  else
  {
    CenterDrag.x = (P1Drag.x+P2Drag.x)/2;
    CenterDrag.y = (P1Drag.y+P2Drag.y)/2;    
  }
  
  P1.x = P1Drag.x*meters_per_pixel;
  P1.y = P1Drag.y*meters_per_pixel;

  P2.x = P2Drag.x*meters_per_pixel;
  P2.y = P2Drag.y*meters_per_pixel;
  
  


  plane_speed = speedSlider.getValue()*mph_to_ms;    // Speed of the plane in MPH 

  // line(100,100,100+2*wind_speed_x,100+2*wind_speed_y);


    turn_rate = turnrateSlider.getValue();      // Time taken for a 360 degree turn, in minutes, standard is 120
  
    // how far will a plane travel in turn_rate seconds? That's the circumfrence of the turning circle.
    base_circumfrence = plane_speed * turn_rate;
    base_radius = base_circumfrence / PI / 2.0f;
    
    // The centerline is the vector from P1 to P2
    centerline = PVector.sub(P1,P2);
    base = new PVector(centerline.y, -centerline.x);
  
    // "angle" is the angle of the centerline, and is basically just added to the anglular postion around the turns
    angle = atan2(base.y, base.x);
  
  
    textAlign(LEFT);
    text("Turn Diameter = "+(base_radius*2/1000)+" km",0,25);
  
    // Caculate various globals  
    // 
    leg_length = centerline.mag();
    leg_time = leg_length/plane_speed;
  
    // Total length of one circuit.  
    // For a racetrack this is just the two legs, plus two half circumfrences
    total_length = (leg_length * 2) + base_circumfrence;
    
    // For a figure 8, it will vary by the angle of the crossing tangents
    
  
    float last_total_time = total_time;
    total_time = (leg_time * 2) + turn_rate;

    base.normalize();
    base.mult(base_radius);

    PVector left_start = new PVector(P1.x + base.x, P1.y + base.y);
    PVector left_end = new PVector(P2.x + base.x, P2.y + base.y);
    PVector right_start = new PVector(P2.x - base.x, P2.y - base.y);
    PVector right_end = new PVector(P1.x - base.x, P1.y - base.y);


  if (!pauseCheckBox.checked())
  {  

  
    if (plane_speed > 0 && last_total_time != total_time && total_time != 1 && last_total_time > 1 && last_total_time < 1000000)
    {
      //    print("fudge");
      time = (time % last_total_time) * total_time / last_total_time;
    } 
    //  print ("last_total_time = "+last_total_time+" total_time"+total_time+" time= = "+time+"\n");
  
  


    ps.update();
  
    if (plane_speed > 0)
    {
      float t = time%total_time;
      time = time+time_step;
  
      // invert the direction
      if (dirCheckBox.checked())
        t = total_time-t;
  
      //  print(t+"\n");
  
      // Now we try to find the point we are at along the track.
  
      if (t < leg_time)
      {
        // the first leg is the "left" stright leg
        //    print("On first leg\n");
        float ratio = t/leg_time;
        PVector fwd;
        
        if (!fig8CheckBox.checked())
        {
           fwd=  PVector.sub(left_end,left_start);          
           fwd.mult(ratio);
           fwd.add(left_start);
        }
        else
        {
            fwd = PVector.sub(left_end,right_end);
            fwd.mult(ratio);
            fwd.add(right_end);
        }

        ps.addParticle(fwd.x/meters_per_pixel,fwd.y/meters_per_pixel);
      }
      else if (t<leg_time+turn_rate/2)
      {
        //    print("On first turn\n");
        float ratio = (t-leg_time)/(turn_rate/2);
  
        // the ratio tells us how far around this turn we are 
        // so we can convert it to 
        float a;
        
          a = angle - PI*ratio;
        
        ps.addParticle((P2.x + base_radius * cos(a) )/meters_per_pixel, (P2.y + base_radius * sin(a))/meters_per_pixel);
      }
      else if (t<leg_time*2+turn_rate/2)
      {
        //    print("On second leg\n");    
        float ratio = (t-leg_time-turn_rate/2)/leg_time;
        PVector fwd = PVector.sub(right_end,right_start);
        
        if (fig8CheckBox.checked())
        {
          fwd = PVector.sub(left_start,right_start);
        }
        
        fwd.mult(ratio);
        fwd.add(right_start);
        ps.addParticle(fwd.x/meters_per_pixel,fwd.y/meters_per_pixel);
      }
      else
      {
        //    print("On second turn\n");
        float ratio = (t-leg_time*2-turn_rate/2)/(turn_rate/2);
  
        // the ratio tells us how far around this turn we are 
        // so we can convert it to 
        float a;

        if (!fig8CheckBox.checked())
          a = PI+angle - PI*ratio;
        else
          a = angle + PI*ratio;

  
        ps.addParticle((P1.x + base_radius * cos(a) )/meters_per_pixel, (P1.y + base_radius * sin(a))/meters_per_pixel);
      }
    }
  }

  noFill();
  stroke(240,240,240);
  ps.render();

  pushMatrix();

  scale(1.0/meters_per_pixel);              // scale down by meters per pixel
  strokeWeight(meters_per_pixel);           // but scale up line thickness, so we can see it.

  // The inner construction lines show up in a dark green
  stroke(0,40,0);
  // Center line
  Pline(P1, P2);

  // Base diameters 
  line(P1.x, P1.y, P1.x + base.x, P1.y + base.y); 
  line(P1.x, P1.y, P1.x - base.x, P1.y - base.y); 

  line(P2.x, P2.y, P2.x + base.x, P2.y + base.y); 
  line(P2.x, P2.y, P2.x - base.x, P2.y - base.y); 

  // The actual track is in a brighter color
  stroke(140,140,0);

  // Left and right legs
  Pline(left_start,left_end);
  Pline(right_start,right_end);

  // Arcs at the ends of the racetrack
  arc(P1.x, P1.y, 2.0*base_radius, 2.0*base_radius,angle, angle+PI);
  arc(P2.x, P2.y, 2.0*base_radius, 2.0*base_radius,angle-PI, angle);


  popMatrix();
  strokeWeight(1.0);
}


void drawGrid()
{
    ADrag.x = 0;
  BDrag.x = 620;
  CDrag.y = 0;
  DDrag.y = 620;

  PVector A = new PVector( ADrag.x*meters_per_pixel, ADrag.y*meters_per_pixel);
  PVector B = new PVector( BDrag.x*meters_per_pixel, BDrag.y*meters_per_pixel);
  PVector C = new PVector( CDrag.x*meters_per_pixel, CDrag.y*meters_per_pixel);
  PVector D = new PVector( DDrag.x*meters_per_pixel, DDrag.y*meters_per_pixel);
  
//  float time_between_planes = 40;
//float time_until_next_plane = 0;  // 0 so we get a plane immediately.
//bool use_AB = true;              // toggle between AB and CD
  
  PVector start, end;
  plane_speed = speedSlider.getValue()*mph_to_ms;    // Speed of the plane in MPH 
  
 time_until_next_plane -= 1;
 if (time_until_next_plane < 0)
 {
   time_until_next_plane = time_between_planes;
   if (use_AB)
   {
     start = A;
     end = B;
   }
   else
   {
     start = C;
     end = D;
   }
   use_AB = !use_AB;
   MPlaneMan.AddPlane(start,end,plane_speed,2000);  // last number is lifetime in ticks
    
   //  MPlaneMan.AddPlane(new PVector(100,100), new PVector(400,600), 1);
   
 }
  
  
  


    MPlaneMan.update();
    MPlaneMan.display();
    
        MUIMan.display();
  
// Draw the AB and CD lines  
  pushMatrix();

  scale(1.0/meters_per_pixel);              // scale down by meters per pixel
  strokeWeight(meters_per_pixel);           // but scale up line thickness, so we can see it.

  stroke(140,140,0);






  Pline(A,B);
  Pline(C,D);


  popMatrix();
  strokeWeight(1.0);
  
  
}

void draw ()
{

  frameRate(30); // Running at 30 is much more consistent.
  
    background(80);

  old_center_x = CenterDrag.x;
  old_center_y = CenterDrag.y;



  image(backgroundImage,0,0);

  textAlign(LEFT);
  text("Contrail Pattern Simulator. By Mick West, contrailscience.com " + calcFPS(),0,13);
  MUIMan.update();
//  MUIMan.display();



  wind_speed = windSlider.getValue()*mph_to_ms;      // Speed of wind in MPH
  wind_angle = headingSlider.getValue()/180*PI;      // Compass bearing that the wind is coming from (0=North, 90 = East, 180=south, 270 = West)
  wind_speed_x = sin(PI+wind_angle)*wind_speed;
  wind_speed_y = -cos(PI+wind_angle)*wind_speed;


  pushMatrix();
  float arrowsize = 6;
  float arrowWeight = 3;
  // Translate to start of arrow
  strokeWeight(arrowWeight);
  translate(100,100);
  stroke(100,100,255);
  // Call vector heading function to get direction (note that pointing up is a heading of 0) and rotate
  rotate(PI+wind_angle);
  // Calculate length of vector & scale it to be bigger or smaller if necessary
  float len = wind_speed;
  if (arrowsize > len/2)
  {
    strokeWeight(arrowWeight*len/2/arrowsize);
    arrowsize = len/2;
  }
  line(0,0,0,-len);
  line(0,-len,-arrowsize/2,-len+arrowsize);
  line(0,-len,arrowsize/2,-len+arrowsize);
  popMatrix();



  if (grid_mode)
  {
    if (!gridCheckBox.checked())
    {
      // Transition to racetrack mode
      P1Drag.show();
      P2Drag.show();
      CenterDrag.show();
      turnrateSlider.show();
      dirCheckBox.show();
      fig8CheckBox.show();  
      
      ADrag.hide();
      BDrag.hide();
      CDrag.hide();
      DDrag.hide();
      
      
      grid_mode = false; 
    }
  }
  else
  {
    if (gridCheckBox.checked())
    {
      
      P1Drag.hide();
      P2Drag.hide();
      CenterDrag.hide();
      turnrateSlider.hide();
      dirCheckBox.hide();
      fig8CheckBox.hide();  
      ADrag.show();
      BDrag.show();
      CDrag.show();
      DDrag.show();
      
      // Transition to grid mode
      grid_mode = true; 
    }
  }
  
  if (grid_mode)
  {
    drawGrid();
  }
  else
  {
    drawRacetrack();
  }
  
}






void displayVector(PVector v, float x, float y, float scayl) {
    pushMatrix();
    float arrowsize = 4;
    // Translate to location to render vector
    translate(x,y);
    stroke(255);
    // Call vector heading function to get direction (note that pointing up is a heading of 0) and rotate
    rotate(v.heading2D());
    // Calculate length of vector & scale it to be bigger or smaller if necessary
    float len = v.mag()*scayl;
    // Draw three lines to make an arrow (draw pointing up since we've rotate to the proper direction)
    line(0,0,len,0);
    line(len,0,len-arrowsize,+arrowsize/2);
    line(len,0,len-arrowsize,-arrowsize/2);
    popMatrix();
  
}

