/* @pjs preload="Modis.jpg"; 
 */

// Contrail simulator
// (c) 2011 Mick West, contrailscience.com

// distances are in meters
// times are in seconds
// speeds are in m/s  (1 mph = 0.44704 m/s)

// Deployment notes:
// Use File -> Export
// This will save to /applet
// Then upload the entire /racetrack_contrail_simulator folder to contrailscience.com/apps

String app_path = "/apps/racetrack_contrail_simulator/";

PImage backgroundImage;

float lifetime = 500;

float start_radius = 1;
float radius_grow = 0.001;

float mph_to_ms = 0.44704;
float miles_to_meters = 1609.344;

float meters_per_pixel = 250.0;

float time_between_planes = 120;
float time_until_next_plane = 0;  // 0 so we get a plane immediately.
boolean use_AB = true;              // toggle between AB and CD


//  PVector P1 = new PVector(400*miles_to_meters,100*miles_to_meters);    // first point
PVector P1 = new PVector(50*miles_to_meters,24*miles_to_meters);    // first point
PVector P2 = new PVector(73*miles_to_meters,25*miles_to_meters);     

float wind_speed_x,wind_speed_y;

float plane_speed, wind_speed, wind_angle, turn_rate, base_circumfrence,base_radius, angle;
PVector centerline,base;
float time = 0;       // current time
float time_step = 10;      // time step in seconds per update fra

// global vars that are calculated in setup
// some bug prevents us using functions like mag in this global initialization

float leg_length;
float total_length;
float leg_time;
float total_time=1;


float  old_center_x;
float  old_center_y;




boolean grid_mode = false; 


MUIElementManager MUIMan;
MPlaneManager MPlaneMan;

MUIVSlider speedSlider,windSlider,headingSlider,turnrateSlider;

MUICheckBox pauseCheckBox, dirCheckBox, fig8CheckBox, gridCheckBox;

MUIDragable P1Drag, P2Drag, CenterDrag;
MUIDragable mark_left_start, mark_left_end, mark_right_start, mark_right_end;


MUIDragable ADrag,BDrag,CDrag,DDrag;


// Helper function to to load an image, accounting for difference in path with processing.js
PImage MUILoadImage(String path)
{
  PImage i = loadImage(path);
  if (i==null || i.width <= 0)
  {
    i = loadImage(app_path+"data/"+path);
  }
  return i;
}

void setup() {
  
  backgroundImage = MUILoadImage("Modis.jpg");

  // Graphics mode setup
  size(640, 640, JAVA2D);
  smooth();
  colorMode(RGB, 255, 255, 255, 100);

  // Setup the particle system
  ps = new ParticleSystem(1000, new PVector(width/2,height/2,0));

  MPlaneMan = new MPlaneManager();

  // MUISlider Manager
  MUIMan = new MUIElementManager();

  speedSlider =      MUIMan.addVSlider(450,400,20,200,100,1000,478,1,"Plane\nSpeed\nMPH");
  windSlider =       MUIMan.addVSlider(500,400,20,200,0,100,44,1,"Wind\nSpeed\nMPH");
  headingSlider =    MUIMan.addVSlider(550,400,20,200,0,360,59,1,"Wind\nHeading\nDeg");
  turnrateSlider =   MUIMan.addVSlider(600,400,20,200,60,1200,360,10,"Turn\nrate\nsecs");
  
  pauseCheckBox = MUIMan.addCheckBox(450,620,20,20,0,"Pause");
  dirCheckBox =   MUIMan.addCheckBox(520,620,20,20,0,"Clockwise");
  fig8CheckBox = MUIMan.addCheckBox(380,620,20,20,0,"Fig8");  
  gridCheckBox = MUIMan.addCheckBox(20,620,20,20,0,"Intersection");
  
  P1Drag = MUIMan.addDragable(50*miles_to_meters/meters_per_pixel,24*miles_to_meters/meters_per_pixel,20,20,0,"P1");
  P2Drag = MUIMan.addDragable(73*miles_to_meters/meters_per_pixel,25*miles_to_meters/meters_per_pixel,20,20,0,"P2");
  
  // AB is a line from the left to the right
  ADrag = MUIMan.addDragable(0,50,20,10,0,"A");  
  BDrag = MUIMan.addDragable(620,150,20,10,0,"B");
  
  // CD is one from the top to the bottom
  CDrag = MUIMan.addDragable(500,0,10,20,0,"C");  
  DDrag = MUIMan.addDragable(150,620,10,20,0,"D");
  
  
  /*
  // some debug stuff
  mark_left_start = MUIMan.addDragable(50,50,10,10,0,"LS");
  mark_left_end = MUIMan.addDragable(0,0,10,10,0,"LE");
  mark_right_start = MUIMan.addDragable(0,0,10,10,0,"RS");
  mark_right_end = MUIMan.addDragable(0,0,10,10,0,"RE");
  */
  
  CenterDrag = MUIMan.addDragable((P1Drag.x+P2Drag.x)/2,(P1Drag.y+P2Drag.y)/2,10,10,0,"");
  
  
  gridCheckBox.value=1;
  
}

void Pline(PVector A, PVector B)
{
  line(A.x, A.y, B.x, B.y);
}

int drawTime = 0;  //for fps
int calcFPS(){
  int diff = millis() - drawTime;
  int fps = (int)(1000/ float(diff));
  drawTime = millis();
  return fps;
} 

void drawRacetrack() {

        MUIMan.display();


  /*  
   slider1.display();
   if(slider1.check(mouseX,mouseY))
   {  
   slider1.move(mouseY);
   // sine1.setAmp(slideramp1.valeur);
   }
   */


  
  if (CenterDrag.x != old_center_x || CenterDrag.y != old_center_y)
  {
    P1Drag.x += CenterDrag.x - old_center_x;
    P1Drag.y += CenterDrag.y - old_center_y;
    P2Drag.x += CenterDrag.x - old_center_x;
    P2Drag.y += CenterDrag.y - old_center_y;
  }
  else
  {
    CenterDrag.x = (P1Drag.x+P2Drag.x)/2;
    CenterDrag.y = (P1Drag.y+P2Drag.y)/2;    
  }
  
  P1.x = P1Drag.x*meters_per_pixel;
  P1.y = P1Drag.y*meters_per_pixel;

  P2.x = P2Drag.x*meters_per_pixel;
  P2.y = P2Drag.y*meters_per_pixel;
  
  


  plane_speed = speedSlider.getValue()*mph_to_ms;    // Speed of the plane in MPH 

  // line(100,100,100+2*wind_speed_x,100+2*wind_speed_y);


    turn_rate = turnrateSlider.getValue();      // Time taken for a 360 degree turn, in minutes, standard is 120
  
    // how far will a plane travel in turn_rate seconds? That's the circumfrence of the turning circle.
    base_circumfrence = plane_speed * turn_rate;
    base_radius = base_circumfrence / PI / 2.0f;
    
    // The centerline is the vector from P1 to P2
    centerline = PVector.sub(P1,P2);
    base = new PVector(centerline.y, -centerline.x);
  
    // "angle" is the angle of the centerline, and is basically just added to the anglular postion around the turns
    angle = atan2(base.y, base.x);
  
  
    textAlign(LEFT);
    text("Turn Diameter = "+(base_radius*2/1000)+" km",0,25);
  
    // Caculate various globals  
    // 
    leg_length = centerline.mag();
    leg_time = leg_length/plane_speed;
  
    // Total length of one circuit.  
    // For a racetrack this is just the two legs, plus two half circumfrences
    total_length = (leg_length * 2) + base_circumfrence;
    
    // For a figure 8, it will vary by the angle of the crossing tangents
    
  
    float last_total_time = total_time;
    total_time = (leg_time * 2) + turn_rate;

    base.normalize();
    base.mult(base_radius);

    PVector left_start = new PVector(P1.x + base.x, P1.y + base.y);
    PVector left_end = new PVector(P2.x + base.x, P2.y + base.y);
    PVector right_start = new PVector(P2.x - base.x, P2.y - base.y);
    PVector right_end = new PVector(P1.x - base.x, P1.y - base.y);


  if (!pauseCheckBox.checked())
  {  

  
    if (plane_speed > 0 && last_total_time != total_time && total_time != 1 && last_total_time > 1 && last_total_time < 1000000)
    {
      //    print("fudge");
      time = (time % last_total_time) * total_time / last_total_time;
    } 
    //  print ("last_total_time = "+last_total_time+" total_time"+total_time+" time= = "+time+"\n");
  
  


    ps.update();
  
    if (plane_speed > 0)
    {
      float t = time%total_time;
      time = time+time_step;
  
      // invert the direction
      if (dirCheckBox.checked())
        t = total_time-t;
  
      //  print(t+"\n");
  
      // Now we try to find the point we are at along the track.
  
      if (t < leg_time)
      {
        // the first leg is the "left" stright leg
        //    print("On first leg\n");
        float ratio = t/leg_time;
        PVector fwd;
        
        if (!fig8CheckBox.checked())
        {
           fwd=  PVector.sub(left_end,left_start);          
           fwd.mult(ratio);
           fwd.add(left_start);
        }
        else
        {
            fwd = PVector.sub(left_end,right_end);
            fwd.mult(ratio);
            fwd.add(right_end);
        }

        ps.addParticle(fwd.x/meters_per_pixel,fwd.y/meters_per_pixel);
      }
      else if (t<leg_time+turn_rate/2)
      {
        //    print("On first turn\n");
        float ratio = (t-leg_time)/(turn_rate/2);
  
        // the ratio tells us how far around this turn we are 
        // so we can convert it to 
        float a;
        
          a = angle - PI*ratio;
        
        ps.addParticle((P2.x + base_radius * cos(a) )/meters_per_pixel, (P2.y + base_radius * sin(a))/meters_per_pixel);
      }
      else if (t<leg_time*2+turn_rate/2)
      {
        //    print("On second leg\n");    
        float ratio = (t-leg_time-turn_rate/2)/leg_time;
        PVector fwd = PVector.sub(right_end,right_start);
        
        if (fig8CheckBox.checked())
        {
          fwd = PVector.sub(left_start,right_start);
        }
        
        fwd.mult(ratio);
        fwd.add(right_start);
        ps.addParticle(fwd.x/meters_per_pixel,fwd.y/meters_per_pixel);
      }
      else
      {
        //    print("On second turn\n");
        float ratio = (t-leg_time*2-turn_rate/2)/(turn_rate/2);
  
        // the ratio tells us how far around this turn we are 
        // so we can convert it to 
        float a;

        if (!fig8CheckBox.checked())
          a = PI+angle - PI*ratio;
        else
          a = angle + PI*ratio;

  
        ps.addParticle((P1.x + base_radius * cos(a) )/meters_per_pixel, (P1.y + base_radius * sin(a))/meters_per_pixel);
      }
    }
  }

  noFill();
  stroke(240,240,240);
  ps.render();

  pushMatrix();

  scale(1.0/meters_per_pixel);              // scale down by meters per pixel
  strokeWeight(meters_per_pixel);           // but scale up line thickness, so we can see it.

  // The inner construction lines show up in a dark green
  stroke(0,40,0);
  // Center line
  Pline(P1, P2);

  // Base diameters 
  line(P1.x, P1.y, P1.x + base.x, P1.y + base.y); 
  line(P1.x, P1.y, P1.x - base.x, P1.y - base.y); 

  line(P2.x, P2.y, P2.x + base.x, P2.y + base.y); 
  line(P2.x, P2.y, P2.x - base.x, P2.y - base.y); 

  // The actual track is in a brighter color
  stroke(140,140,0);

  // Left and right legs
  Pline(left_start,left_end);
  Pline(right_start,right_end);

  // Arcs at the ends of the racetrack
  arc(P1.x, P1.y, 2.0*base_radius, 2.0*base_radius,angle, angle+PI);
  arc(P2.x, P2.y, 2.0*base_radius, 2.0*base_radius,angle-PI, angle);


  popMatrix();
  strokeWeight(1.0);
}


void drawGrid()
{
    ADrag.x = 0;
  BDrag.x = 620;
  CDrag.y = 0;
  DDrag.y = 620;

  PVector A = new PVector( ADrag.x*meters_per_pixel, ADrag.y*meters_per_pixel);
  PVector B = new PVector( BDrag.x*meters_per_pixel, BDrag.y*meters_per_pixel);
  PVector C = new PVector( CDrag.x*meters_per_pixel, CDrag.y*meters_per_pixel);
  PVector D = new PVector( DDrag.x*meters_per_pixel, DDrag.y*meters_per_pixel);
  
//  float time_between_planes = 40;
//float time_until_next_plane = 0;  // 0 so we get a plane immediately.
//bool use_AB = true;              // toggle between AB and CD
  
  PVector start, end;
  plane_speed = speedSlider.getValue()*mph_to_ms;    // Speed of the plane in MPH 
  
 time_until_next_plane -= 1;
 if (time_until_next_plane < 0)
 {
   time_until_next_plane = time_between_planes;
   if (use_AB)
   {
     start = A;
     end = B;
   }
   else
   {
     start = C;
     end = D;
   }
   use_AB = !use_AB;
   MPlaneMan.AddPlane(start,end,plane_speed,2000);  // last number is lifetime in ticks
    
   //  MPlaneMan.AddPlane(new PVector(100,100), new PVector(400,600), 1);
   
 }
  
  
  


    MPlaneMan.update();
    MPlaneMan.display();
    
        MUIMan.display();
  
// Draw the AB and CD lines  
  pushMatrix();

  scale(1.0/meters_per_pixel);              // scale down by meters per pixel
  strokeWeight(meters_per_pixel);           // but scale up line thickness, so we can see it.

  stroke(140,140,0);






  Pline(A,B);
  Pline(C,D);


  popMatrix();
  strokeWeight(1.0);
  
  
}

void draw ()
{

  frameRate(30); // Running at 30 is much more consistent.
  
    background(80);

  old_center_x = CenterDrag.x;
  old_center_y = CenterDrag.y;



  image(backgroundImage,0,0);

  textAlign(LEFT);
  text("Contrail Pattern Simulator. By Mick West, contrailscience.com " + calcFPS(),0,13);
  MUIMan.update();
//  MUIMan.display();



  wind_speed = windSlider.getValue()*mph_to_ms;      // Speed of wind in MPH
  wind_angle = headingSlider.getValue()/180*PI;      // Compass bearing that the wind is coming from (0=North, 90 = East, 180=south, 270 = West)
  wind_speed_x = sin(PI+wind_angle)*wind_speed;
  wind_speed_y = -cos(PI+wind_angle)*wind_speed;


  pushMatrix();
  float arrowsize = 6;
  float arrowWeight = 3;
  // Translate to start of arrow
  strokeWeight(arrowWeight);
  translate(100,100);
  stroke(100,100,255);
  // Call vector heading function to get direction (note that pointing up is a heading of 0) and rotate
  rotate(PI+wind_angle);
  // Calculate length of vector & scale it to be bigger or smaller if necessary
  float len = wind_speed;
  if (arrowsize > len/2)
  {
    strokeWeight(arrowWeight*len/2/arrowsize);
    arrowsize = len/2;
  }
  line(0,0,0,-len);
  line(0,-len,-arrowsize/2,-len+arrowsize);
  line(0,-len,arrowsize/2,-len+arrowsize);
  popMatrix();



  if (grid_mode)
  {
    if (!gridCheckBox.checked())
    {
      // Transition to racetrack mode
      P1Drag.show();
      P2Drag.show();
      CenterDrag.show();
      turnrateSlider.show();
      dirCheckBox.show();
      fig8CheckBox.show();  
      
      ADrag.hide();
      BDrag.hide();
      CDrag.hide();
      DDrag.hide();
      
      
      grid_mode = false; 
    }
  }
  else
  {
    if (gridCheckBox.checked())
    {
      
      P1Drag.hide();
      P2Drag.hide();
      CenterDrag.hide();
      turnrateSlider.hide();
      dirCheckBox.hide();
      fig8CheckBox.hide();  
      ADrag.show();
      BDrag.show();
      CDrag.show();
      DDrag.show();
      
      // Transition to grid mode
      grid_mode = true; 
    }
  }
  
  if (grid_mode)
  {
    drawGrid();
  }
  else
  {
    drawRacetrack();
  }
  
}






void displayVector(PVector v, float x, float y, float scayl) {
    pushMatrix();
    float arrowsize = 4;
    // Translate to location to render vector
    translate(x,y);
    stroke(255);
    // Call vector heading function to get direction (note that pointing up is a heading of 0) and rotate
    rotate(v.heading2D());
    // Calculate length of vector & scale it to be bigger or smaller if necessary
    float len = v.mag()*scayl;
    // Draw three lines to make an arrow (draw pointing up since we've rotate to the proper direction)
    line(0,0,len,0);
    line(len,0,len-arrowsize,+arrowsize/2);
    line(len,0,len-arrowsize,-arrowsize/2);
    popMatrix();
  
}

/*
 MickPlane.pde
 Processing 2D Plane Simulation
 
 Released under The MIT License
 
 Copyright (c) 2012 Mick West
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 
 */
 
 class MPlane
 {
   
   PVector pos;
   PVector vel;
   
   boolean alive = true;
   
   ParticleSystem ps;
   
   int next_particle;
   int particle_gap = 10;
   
   int lifetime;
   
   
   
   MPlane(PVector _pos, PVector _vel, int _lifetime)
   {
     pos = _pos;
     vel = _vel;
     ps = new ParticleSystem(100, new PVector(width/2,height/2,0));
     next_particle = 0;
     lifetime = _lifetime;
   }
   
   
    void display()
    {
      float x = pos.x/meters_per_pixel;
      float y = pos.y/meters_per_pixel;
      float w=10;
      float h=10;
    
      stroke(255);
      strokeWeight(2);
      line(x-w/2,y-h/2,x+w/2,y+h/2);
      line(x-w/2,y+h/2,x+w/2,y-h/2);
      
      next_particle -= 1;
      if (next_particle <= 0)
      {
        next_particle = particle_gap;
        ps.addParticle(x,y);
      }
      
      
      ps.update();
      ps.render();
   }
   
  void update()
  {
    pos.x = pos.x + vel.x * time_step; 
    pos.y = pos.y + vel.y * time_step;
    lifetime -= 1;
    if (lifetime < 0)
    {
      die();
    }
 
  } 
   
  boolean dead()
  {
    return !alive;
  }

  void die()
  {
    alive = false;
  }
   
 }
 
 class MPlaneManager
 {
   ArrayList planes;
   
   MPlaneManager()
   {
     planes = new ArrayList();
   }
   
   MPlane AddPlane(PVector A, PVector B, float speed, int lifetime)
   {
     PVector AB = PVector.sub(B,A);
     
     PVector start = PVector.sub(A,AB);
         
     PVector vel = AB;
     vel.normalize();
     
     vel.mult(speed);
     
     // start at A - AB, so we get the off-screen portion of the contrail
     MPlane p = new MPlane(start,vel,lifetime);
     planes.add(p);
     return p;  
   }
   
   void update()
   {
     
      for (int i = planes.size()-1; i >= 0; i--) 
      {
        MPlane p = (MPlane) planes.get(i);
          p.update();
      }
   prune();
     
   }
   
   void display()
   {
      for (int i = planes.size()-1; i >= 0; i--) 
      {
        MPlane p = (MPlane) planes.get(i);
          p.display();
      }
   
   }
   
   
  void prune()
  {
    for (int i = planes.size()-1; i >= 0; i--) 
    {
      MPlane p = (MPlane) planes.get(i);
      if (p.dead()) {
        planes.remove(i);
      }
    }
  }
   
 }
/*
 Processing User Interface Elements
 Compatible with Processing.js (some default font differences apply, and vaires with browser)
 
 Released under The MIT License
 
 Copyright (c) 2011 Mick West
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 
 */
 
 /*
 
 Example usage:

// Variables, these can be globals for simplicity. 

// At a minimum you need one manager object. This handles updating things, and drawing them.
MUIElementManager MUIMan;

// Variables for sliders
MUIVSlider speedSlider,windSlider,headingSlider,turnrateSlider;

// Variables for Check Boxes
MUICheckBox pauseCheckBox, dirCheckBox, fig8CheckBox;

// Varaible for dragable boxes
MUIDragable P1Drag, P2Drag, CenterDrag;

// INITIALIZATION

// This would be somethwere in your setup code. You need to create the Element manager

  // MUISlider Manager
  MUIMan = new MUIElementManager();

// Then add elements
// Vslider are vertical sliders. 
//  parameters are (float x_, float y_, float w_, float h_, float start_, float end_, float initial_, float _resolution, String _lable)

  speedSlider =      MUIMan.addVSlider(450,400,10,200,100,1000,478,1,"Plane\nSpeed\nMPH");
  windSlider =       MUIMan.addVSlider(500,400,10,200,0,100,44,1,"Wind\nSpeed\nMPH");
  headingSlider =    MUIMan.addVSlider(550,400,10,200,0,360,59,1,"Wind\nHeading\nDeg");
  turnrateSlider =   MUIMan.addVSlider(600,400,10,200,60,1200,360,10,"Turn\nrate\nsecs");
  
  pauseCheckBox = MUIMan.addCheckBox(450,620,10,10,0,"Pause");
  dirCheckBox =   MUIMan.addCheckBox(520,620,10,10,0,"Clockwise");
  fig8CheckBox = MUIMan.addCheckBox(380,620,10,10,0,"Fig8");
  
  P1Drag = MUIMan.addDragable(50*miles_to_meters/meters_per_pixel,24*miles_to_meters/meters_per_pixel,10,10,0,"P1");
  P2Drag = MUIMan.addDragable(73*miles_to_meters/meters_per_pixel,25*miles_to_meters/meters_per_pixel,10,10,0,"P2");
  
  CenterDrag = MUIMan.addDragable((P1Drag.x+P2Drag.x)/2,(P1Drag.y+P2Drag.y)/2,10,10,0,"");
 
 Then you just need to call the "update" and "display" functions
 
  MUIMan.update();
  MUIMan.display();
  
  The UI elements will be displayed, and you can then access their state (for sliders and check boxes) and position (for draggable points).
   
 */
 

class MUIElement
{
  float x,y;                    // position of the center of the element in screen space  
  float value;                  // value of the slider calculated 
  float w,h;                    // width and height (not currently the actual width and height for sliders, the border is added) 
  String  lable;                // lable below the slider. Can have line feeds, like  "max\nspeed\nin MPH"
  boolean alive = true;         // flag if alive. setting it to false will disable it, and it will be pruned by the slider manager
  float border_x = 0;          // border to left and right of the slider
  float border_y = 0;          // border above and below
  
  boolean visible = true;     // flag if it is displayable


  void display() {
  };
  void update() {
  };

  boolean dead()
  {
    return !alive;
  }

  void die()
  {
    alive = false;
  }

  void hide()
  {
    visible = false;
  }
  
  void show()
  {
    visible = true;
  }
  
  // handle a mouse click
  // returns true if we've handled it, and don't want to propogate
  boolean  handleClick(float x, float y) {  return false;  }
  boolean  handleRelease(float x, float y) {  return false;  }
  boolean  handleMove(float x, float y) {  return false;  }
  
}

class MUIVSlider extends MUIElement
{
  float yp;                     // y position of the slider bar in screen space
  float start, end;             // start (slider all the way down) and end (slider all the way up) values
  float resolution;             // the smallest step that the slider value is output in.

  float button_height = 20;     // height of the slider button, w is used as the width of the button 
  boolean dragging = false;
 
  // Constructor, generally this will be called by a slider manager
  MUIVSlider(float x_, float y_, float w_, float h_, float start_, float end_, float initial_, float _resolution, String _lable)
  {
    x = x_;
    y = y_;
    w = w_;
    h = h_;

    resolution = _resolution;

    start=start_;
    end=end_;

    lable = _lable;

    value = (initial_ - start) / (end - start);
    yp = y-h/2+h*(1-value);
    
    border_x = 10;
    border_y = 10;
    
  }


  void display()
  {
    rectMode(CENTER);

    // Background rectangle
    fill(40);
    noStroke();
    rect(x,y,w+border_x,h+button_height+border_y);

    stroke(255);
    strokeWeight(4);
    line(x,y-h/2,x,y+h/2); // The center line

    strokeWeight(1);
    fill(128); 
    rect(x,yp,w,button_height);
    strokeWeight(3);
    line(x+w/2,yp,x-w/2,yp);

    pushMatrix();    
    float  s = 1;
    scale (s);
    fill(0,255,0,255);
    textAlign(CENTER);
    // processing.js does not support the extra width and height parameters
    // so we just use the x,y version, which works fine with a little tweaking.
    text(""+getValue(),x/s,(y-h/2-border_y-5*s)/s);   // Note the ""+getValue() is used to convert the getValue to a String
    text(lable,x/s,(y+h/2+border_y+13*s)/s);
    popMatrix();  // matrix was pushed for the scale
  }
  
  boolean  handleClick(float mx, float my)
  {
    if((mx>x-(w/2+border_x) && mx<x+(w/2+border_x) && my>y-(h/2+button_height+border_y) && my<y+(h/2+button_height+border_y)) )
    {
      dragging = true;
      handleMove(mx,my);
      return true;  
    }
    else
    {
      return false;
    }    
  }
  
  
  boolean handleRelease(float mx, float my)
  {
    dragging = false;
    return false;  
  }
  
  boolean handleMove(float mx, float my)
  {
    if (dragging)
    {
      yp=my;
      yp=constrain(yp,y-h/2,y+h/2);  
      value = map(yp,y+h/2,y-h/2,0,1);
      value = constrain(value,0,1);
      return true;
    }
    else
    {
      return false;
    }
  }
  
  


  float getValue()
  {
    float v = start + (end-start)*value;
    // scale to nearest resolution
    v = v - (v % resolution);
    return v;
  }
}


class MUICheckBox extends MUIElement
{

  MUICheckBox(float _x, float _y, float _w, float _h, float _value, String _lable)
  {
    x = _x;
    y = _y;
    w = _w;
    h = _h;
    value = _value;
    lable = _lable;
    
    border_x = 5;
    border_y = 5;
    
  }
  
  void display()
  {
    strokeWeight(2);
    rectMode(CENTER);
    fill(0);
    noStroke();
    rect(x,y,w,h);
    stroke(128);
    noFill();
    rect(x,y,w,h);
    
    strokeWeight(2);
    if (value != 0)
    {
      line(x-w/2,y-h/2,x+w/2,y+h/2);
      line(x-w/2,y+h/2,x+w/2,y-h/2);
    } 

    stroke(0,255,0);
    fill(0,255,0,255);
    text(lable,x+w+5,y+4);


/*    
    if (mousePressed)
    {
      strokeWeight(1);
      line(mouseX-w/2,mouseY-h/2,mouseX+w/2,mouseY+h/2);
      line(mouseX-w/2,mouseY+h/2,mouseX+w/2,mouseY-h/2);
    }
*/    
  }
  
  /*
  void update()
  {
    if(mousePressed && (mouseX>x-(w/2+border_x) && mouseX<x+(w/2+border_x) && mouseY>y-(h/2+border_y) && mouseY<y+(h/2+border_y)) )
    {
      value = (value == 0)?1:0;  
    }
  }
  */
  
  boolean  handleClick(float mx, float my)
  {
    if((mx > x-(w/2+border_x) && mx < x+(w/2+border_x) && my > y-(h/2+border_y) && my < y+(h/2+border_y)) )
    {
      value = (value == 0)?1:0;
      return true;  
    }
    else
    {
      return false;
    }
    
  }
  
  boolean checked()
  {
    return value != 0;
  }
  
}

class MUIDragable extends MUIElement
{
  
  boolean dragging;
  float offset_x, offset_y;

  MUIDragable(float _x, float _y, float _w, float _h, float _value, String _lable)
  {
    x = _x;
    y = _y;
    w = _w;
    h = _h;
    value = _value;
    lable = _lable;
    
    border_x = 5;
    border_y = 5;

    dragging = false;
    
  }
  
  void display()
  {
    strokeWeight(2);
    rectMode(CENTER);
    fill(0);
    noStroke();
    rect(x,y,w,h);
    stroke(128);
    noFill();
    rect(x,y,w,h);
    
    strokeWeight(2);
    if (value != 0)
    {
      line(x-w/2,y-h/2,x+w/2,y+h/2);
      line(x-w/2,y+h/2,x+w/2,y-h/2);
    } 

    stroke(0,255,0);
    fill(0,255,0,255);
    text(lable,x+w+5,y+4);
  }
  
  boolean  handleClick(float mx, float my)
  {
    if((mx > x-(w/2+border_x) && mx < x+(w/2+border_x) && my > y-(h/2+border_y) && my < y+(h/2+border_y)) )
    {
      offset_x = mx-x;
      offset_y = my-y;
      dragging = true;
      return true;  
    }
    else
    {
      return false;
    }    
  }
  
  
  boolean handleRelease(float mx, float my)
  {
    dragging = false;
    return false;  
  }
  
  boolean handleMove(float mx, float my)
  {
    if (dragging)
    {
      x = mx-offset_x;
      y = my-offset_y;
      return true;
    }
    else
    {
      return false;
    }
  }
  
}


// The slider manger class simply mantains a list of the sliders
// and runs the update and display functions on them, taking care of 
// removing any dead sliders.
class MUIElementManager
{
  ArrayList elements;

  boolean mouse_pressed   = false;
  boolean mouse_clicked   = false;
  boolean mouse_released  = false;
  boolean mouse_moved;
  float mouse_x = 0;
  float mouse_y = 0;  

  MUIElementManager()
  {
    elements = new ArrayList();
  }

  

  void update()
  {

    if (mouse_x != mouseX || mouse_y != mouseY)
    {
      mouse_moved = true;
    }

    mouse_x = mouseX;
    mouse_y = mouseY;
    
    if (mousePressed)
    {
      if (mouse_pressed == false)
      {
        mouse_clicked = true;
      }
      mouse_pressed = true;
    }
    else
    {
      if (mouse_pressed)
      {
        mouse_released = true;
      }
      // this is a bit long winded, but I'm doing it to avoid reading mousePressed twice, which could 
      // screw up the logic if it's updated asyn
      mouse_pressed = false;
    }
    
    for (int i = elements.size()-1; i >= 0; i--) {
      MUIElement p = (MUIElement) elements.get(i);
      if (!p.dead())
      {
        if (mouse_clicked)
        {
          if (p.handleClick(mouse_x, mouse_y))
          {
            mouse_clicked = false;
          }
        }
      

        if (mouse_released)
        {
          if (p.handleRelease(mouse_x, mouse_y))
          {
            mouse_released = false;
          }
        }
        

        if (mouse_moved)
        {
          if (p.handleMove(mouse_x, mouse_y))
          {
            mouse_moved = false;
          }
        }
        

        if (!p.dead())  // click handledling might have killed it
        {
          p.update();
        }
      }
    }        
    mouse_clicked = false;  // if handled the click, then drop it    
    mouse_released = false;  // if handled the click, then drop it    
    mouse_moved = false;  // if handled the click, then drop it    
    prune();
  }

  void display()
  {
    for (int i = elements.size()-1; i >= 0; i--) {
      MUIElement p = (MUIElement) elements.get(i);
      if (!p.dead() && p.visible)
      {
        p.display();
      }
    }
  }

  void prune()
  {
    for (int i = elements.size()-1; i >= 0; i--) 
    {
      MUIElement p = (MUIElement) elements.get(i);
      if (p.dead()) {
        elements.remove(i);
      }
    }
  }

// Utility Functions, you could just do 
//   MUIMan.add(new MUIVSlider(...));
// but this lets you do:
//   MUIMan.addVSlider(...)
// which is a bit nicer, and encapsulates things slightly if I need to change the registering method below

  MUIVSlider addVSlider(float _x, float _y, float _w, float _h, float _start, float _end, float _initial, float _resolution, String lable_)
  {
    MUIVSlider s = new MUIVSlider(_x,_y,_w,_h,_start,_end,_initial,_resolution, lable_);
    elements.add(s);
    return s;
  }

  MUICheckBox addCheckBox(float _x, float _y, float _w, float _h, float _value, String _lable)
  {
    MUICheckBox s = new MUICheckBox(_x, _y, _w, _h, _value,_lable);
    elements.add(s);
    return s;
  }

  MUIDragable addDragable(float _x, float _y, float _w, float _h, float _value, String _lable)
  {
    MUIDragable s = new MUIDragable(_x, _y, _w, _h, _value,_lable);
    elements.add(s);
    return s;
  }


  void remove(MUIElement s)
  {
    // Mark it as dead
    s.die();
    prune();
  }
}

ParticleSystem ps;


// A class to describe a group of Particles
// An ArrayList is used to manage the list of Particles 

class ParticleSystem {

  Particle[] particles;

  int num_slots;         // total slots
  int num_particles;      // how many are used
  int head;              // the newest used slot
  int tail;              // the oldest used slot

  PVector origin;        // An origin point for where particles are born

    ParticleSystem(int num, PVector v) {
    particles = new Particle[num];
    origin = v.get();                        // Store the origin point
    num_slots = num;
    num_particles = 0;
    head = 0;
    tail = 0;
  }

  void update() 
  {
    int p = head;
    for (int i = 0; i < num_particles; i++ ) {
      particles[p].update(); 
      p--;   
      if (p < 0) p += num_slots;
    }
  }

  void render()
  {
    int p = head;
    for (int i = 0; i < num_particles-1; i++ ) {
      int next = p-1;
      if (next < 0) next += num_slots; 
      if (p != head)
      {
        ellipseMode(CENTER);
        //    stroke(255,timer);
        //    fill(100,timer);
        //    ellipse(particles[p].loc.x,particles[p].loc.y,particles[p].r,particles[p].r);
        strokeWeight(particles[p].r*2);
        //     float alp = particles[p].r/10;
        //     if (alp > 1.0) alp = 1.0;
        //     stroke(#FFFFFF,255);
        line(particles[p].loc.x,particles[p].loc.y,particles[next].loc.x,particles[next].loc.y);
      }
      p=next;
    }
  }

  void addParticle() {
    //   particles.add(new Particle(origin));
  }

  void addParticle(float x, float y) {

    // increment the head to the next slot
    if (num_particles != 0)
    {
      head++;
      // wrap as needed
      if (head == num_slots)
      {
        head = 0;
      }
    }

    particles[head] = (new Particle(new PVector(x,y)));

    // if it overrides the tail, then advance the tail
    if (num_particles == num_slots)
    {
      tail++;
      if (tail == num_slots)
      {
        tail = 0;
      }
    }
    else
    {
      num_particles++;
    }
  }

  void addParticle(Particle p) {
    //   particles.add(p);
  }

  // A method to test if the particle system still has particles
  boolean dead() {
    //   if (particles.isEmpty()) {
    //     return true;
    //   } else {
    return false;
    //   }
  }
};


// A simple Particle class

class Particle {
  PVector loc;
  PVector vel;
  PVector acc;
  float r;
  float timer;

  // Another constructor (the one we are using here)
  Particle(PVector l) {
    //   acc = new PVector(0,0.05,0);
    //   vel = new PVector(random(-1,1),random(-2,0),0);

    acc = new PVector(0,0,0);
    //    vel = new PVector(-wind_speed*time_step/meters_per_pixel,wind_speed*time_step/meters_per_pixel,0);
    vel = new PVector(0,0,0);  

    loc = l.get();
    r = start_radius;
    timer = lifetime;
  }

  // Method to update location
  void update() {
    vel.add(acc);
    loc.x += wind_speed_x*time_step/meters_per_pixel;
    loc.y += wind_speed_y*time_step/meters_per_pixel;

    loc.add(vel);
    timer -= 1.0;
    r += radius_grow * time_step;
  }

  // Method to display
  void render() {
    ellipseMode(CENTER);
    stroke(255,timer);
    fill(100,timer);
    ellipse(loc.x,loc.y,r,r);
    //   displayVector(vel,loc.x,loc.y,10);
  }

  // Is the particle still useful?
  boolean dead() {
    if (timer <= 0.0) {
      return true;
    } 
    else {
      return false;
    }
  }
}

